@@ -96,7 +96,7 @@ type Breaklet struct {
 	// the return value will determine if the breaklet should be considered
 	// active.
 	// The callback can have side-effects.
-	callback func(th Thread) bool
+	callback func(th Thread, p *Target) (bool, error)
 
 	// For WatchOutOfScopeBreakpoints and StackResizeBreakpoints the watchpoint
 	// field contains the watchpoint related to this out of scope sentinel.
@@ -131,6 +131,10 @@ const (
 	// adjust the watchpoint of stack variables.
 	StackResizeBreakpoint
 
+	// PluginOpenBreakpoint is a breakpoint used to detect that a plugin has
+	// been loaded and we should try to enable suspended breakpoints.
+	PluginOpenBreakpoint
+
 	steppingMask = NextBreakpoint | NextDeferBreakpoint | StepBreakpoint
 )
 
@@ -204,6 +208,8 @@ func (bp *Breakpoint) VerboseDescr() []string {
 			r = append(r, fmt.Sprintf("WatchOutOfScope Cond=%q checkPanicCall=%v", exprToString(breaklet.Cond), breaklet.checkPanicCall))
 		case StackResizeBreakpoint:
 			r = append(r, fmt.Sprintf("StackResizeBreakpoint Cond=%q", exprToString(breaklet.Cond)))
+		case PluginOpenBreakpoint:
+			r = append(r, "PluginOpenBreakpoint")
 		default:
 			r = append(r, fmt.Sprintf("Unknown %d", breaklet.Kind))
 		}
@@ -264,20 +270,22 @@ func (bpstate *BreakpointState) checkCond(tgt *Target, breaklet *Breaklet, threa
 
 	switch breaklet.Kind {
 	case UserBreakpoint:
+		var goroutineID int64
 		lbp := bpstate.Breakpoint.Logical
 		if lbp != nil {
 			if g, err := GetG(thread); err == nil {
-				lbp.HitCount[g.ID]++
+				goroutineID = g.ID
+				lbp.HitCount[goroutineID]++
 			}
 			lbp.TotalHitCount++
 		}
-		active = checkHitCond(lbp)
+		active = checkHitCond(lbp, goroutineID)
 
 	case StepBreakpoint, NextBreakpoint, NextDeferBreakpoint:
 		nextDeferOk := true
 		if breaklet.Kind&NextDeferBreakpoint != 0 {
 			var err error
-			frames, err := ThreadStacktrace(thread, 2)
+			frames, err := ThreadStacktrace(tgt, thread, 2)
 			if err == nil {
 				nextDeferOk, _ = isPanicCall(frames)
 				if !nextDeferOk {
@@ -286,23 +294,17 @@ func (bpstate *BreakpointState) checkCond(tgt *Target, breaklet *Breaklet, threa
 			}
 		}
 		active = active && nextDeferOk
-		if active {
-			bpstate.Stepping = true
-			if breaklet.Kind == StepBreakpoint {
-				bpstate.SteppingInto = true
-			}
-		}
 
 	case WatchOutOfScopeBreakpoint:
 		if breaklet.checkPanicCall {
-			frames, err := ThreadStacktrace(thread, 2)
+			frames, err := ThreadStacktrace(tgt, thread, 2)
 			if err == nil {
 				ipc, _ := isPanicCall(frames)
 				active = active && ipc
 			}
 		}
 
-	case StackResizeBreakpoint:
+	case StackResizeBreakpoint, PluginOpenBreakpoint:
 		// no further checks
 
 	default:
@@ -311,33 +313,51 @@ func (bpstate *BreakpointState) checkCond(tgt *Target, breaklet *Breaklet, threa
 
 	if active {
 		if breaklet.callback != nil {
-			active = breaklet.callback(thread)
+			var err error
+			active, err = breaklet.callback(thread, tgt)
+			if err != nil && bpstate.CondError == nil {
+				bpstate.CondError = err
+			}
 		}
 		bpstate.Active = active
 	}
+
+	if bpstate.Active {
+		switch breaklet.Kind {
+		case NextBreakpoint, NextDeferBreakpoint:
+			bpstate.Stepping = true
+		case StepBreakpoint:
+			bpstate.Stepping = true
+			bpstate.SteppingInto = true
+		}
+	}
 }
 
 // checkHitCond evaluates bp's hit condition on thread.
-func checkHitCond(lbp *LogicalBreakpoint) bool {
+func checkHitCond(lbp *LogicalBreakpoint, goroutineID int64) bool {
 	if lbp == nil || lbp.HitCond == nil {
 		return true
 	}
+	hitCount := int(lbp.TotalHitCount)
+	if lbp.HitCondPerG && goroutineID > 0 {
+		hitCount = int(lbp.HitCount[goroutineID])
+	}
 	// Evaluate the breakpoint condition.
 	switch lbp.HitCond.Op {
 	case token.EQL:
-		return int(lbp.TotalHitCount) == lbp.HitCond.Val
+		return hitCount == lbp.HitCond.Val
 	case token.NEQ:
-		return int(lbp.TotalHitCount) != lbp.HitCond.Val
+		return hitCount != lbp.HitCond.Val
 	case token.GTR:
-		return int(lbp.TotalHitCount) > lbp.HitCond.Val
+		return hitCount > lbp.HitCond.Val
 	case token.LSS:
-		return int(lbp.TotalHitCount) < lbp.HitCond.Val
+		return hitCount < lbp.HitCond.Val
 	case token.GEQ:
-		return int(lbp.TotalHitCount) >= lbp.HitCond.Val
+		return hitCount >= lbp.HitCond.Val
 	case token.LEQ:
-		return int(lbp.TotalHitCount) <= lbp.HitCond.Val
+		return hitCount <= lbp.HitCond.Val
 	case token.REM:
-		return int(lbp.TotalHitCount)%lbp.HitCond.Val == 0
+		return hitCount%lbp.HitCond.Val == 0
 	}
 	return false
 }
@@ -534,7 +554,7 @@ func (t *Target) setEBPFTracepointOnFunc(fn *Function, goidOffset int64) error {
 	if t.BinInfo().Producer() != "" && goversion.ProducerAfterOrEqual(t.BinInfo().Producer(), 1, 15) {
 		variablesFlags |= reader.VariablesTrustDeclLine
 	}
-	_, l, _ := t.BinInfo().PCToLine(fn.Entry)
+	_, l := t.BinInfo().EntryLineForFunc(fn)
 
 	var args []ebpf.UProbeArgMap
 	varEntries := reader.Variables(dwarfTree, fn.Entry, l, variablesFlags)
@@ -569,8 +589,7 @@ func (t *Target) setEBPFTracepointOnFunc(fn *Function, goidOffset int64) error {
 	//TODO(aarzilli): inlined calls?
 
 	// Finally, set the uprobe on the function.
-	t.proc.SetUProbe(fn.Name, goidOffset, args)
-	return nil
+	return t.proc.SetUProbe(fn.Name, goidOffset, args)
 }
 
 // SetWatchpoint sets a data breakpoint at addr and stores it in the
@@ -633,7 +652,7 @@ func (t *Target) SetWatchpoint(logicalID int, scope *EvalScope, expr string, wty
 
 func (t *Target) setBreakpointInternal(logicalID int, addr uint64, kind BreakpointKind, wtype WatchType, cond ast.Expr) (*Breakpoint, error) {
 	if valid, err := t.Valid(); !valid {
-		recorded, _ := t.Recorded()
+		recorded, _ := t.recman.Recorded()
 		if !recorded {
 			return nil, err
 		}
@@ -654,7 +673,7 @@ func (t *Target) setBreakpointInternal(logicalID int, addr uint64, kind Breakpoi
 		lbp := bpmap.Logical[logicalID]
 		if lbp == nil {
 			lbp = &LogicalBreakpoint{LogicalID: logicalID}
-			lbp.HitCount = make(map[int]uint64)
+			lbp.HitCount = make(map[int64]uint64)
 			lbp.Enabled = true
 			bpmap.Logical[logicalID] = lbp
 		}
@@ -663,8 +682,13 @@ func (t *Target) setBreakpointInternal(logicalID int, addr uint64, kind Breakpoi
 		if breaklet != nil && breaklet.Cond == nil {
 			breaklet.Cond = lbp.Cond
 		}
-		lbp.File = bp.File
-		lbp.Line = bp.Line
+		if lbp.File == "" && lbp.Line == 0 {
+			lbp.File = bp.File
+			lbp.Line = bp.Line
+		} else if bp.File != lbp.File || bp.Line != lbp.Line {
+			lbp.File = "<multiple locations>"
+			lbp.Line = 0
+		}
 		fn := t.BinInfo().PCToFunc(bp.Addr)
 		if fn != nil {
 			lbp.FunctionName = fn.NameWithoutTypeParams()
@@ -738,7 +762,7 @@ func (bp *Breakpoint) canOverlap(kind BreakpointKind) bool {
 // ClearBreakpoint clears the breakpoint at addr.
 func (t *Target) ClearBreakpoint(addr uint64) error {
 	if valid, err := t.Valid(); !valid {
-		recorded, _ := t.Recorded()
+		recorded, _ := t.recman.Recorded()
 		if !recorded {
 			return err
 		}
@@ -961,6 +985,8 @@ type LogicalBreakpoint struct {
 	Line         int
 	Enabled      bool
 
+	Set SetBreakpoint
+
 	Tracepoint  bool // Tracepoint flag
 	TraceReturn bool
 	Goroutine   bool     // Retrieve goroutine information
@@ -969,8 +995,9 @@ type LogicalBreakpoint struct {
 	LoadArgs    *LoadConfig
 	LoadLocals  *LoadConfig
 
-	HitCount      map[int]uint64 // Number of times a breakpoint has been reached in a certain goroutine
-	TotalHitCount uint64         // Number of times a breakpoint has been reached
+	HitCount      map[int64]uint64 // Number of times a breakpoint has been reached in a certain goroutine
+	TotalHitCount uint64           // Number of times a breakpoint has been reached
+	HitCondPerG   bool             // Use per goroutine hitcount as HitCond operand, instead of total hitcount
 
 	// HitCond: if not nil the breakpoint will be triggered only if the evaluated HitCond returns
 	// true with the TotalHitCount.
@@ -984,3 +1011,18 @@ type LogicalBreakpoint struct {
 
 	UserData interface{} // Any additional information about the breakpoint
 }
+
+// SetBreakpoint describes how a breakpoint should be set.
+type SetBreakpoint struct {
+	FunctionName string
+	File         string
+	Line         int
+	Expr         func(*Target) []uint64
+	ExprString   string
+	PidAddrs     []PidAddr
+}
+
+type PidAddr struct {
+	Pid  int
+	Addr uint64
+}
