@@ -3,9 +3,9 @@ package native
 import (
 	"bufio"
 	"bytes"
+	"debug/elf"
 	"errors"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"os/exec"
 	"os/signal"
@@ -18,6 +18,7 @@ import (
 
 	sys "golang.org/x/sys/unix"
 
+	"github.com/go-delve/delve/pkg/logflags"
 	"github.com/go-delve/delve/pkg/proc"
 	"github.com/go-delve/delve/pkg/proc/internal/ebpf"
 	"github.com/go-delve/delve/pkg/proc/linutil"
@@ -61,7 +62,7 @@ func (os *osProcessDetails) Close() {
 // to be supplied to that process. `wd` is working directory of the program.
 // If the DWARF information cannot be found in the binary, Delve will look
 // for external debug files in the directories passed in.
-func Launch(cmd []string, wd string, flags proc.LaunchFlags, debugInfoDirs []string, tty string, redirects [3]string) (*proc.Target, error) {
+func Launch(cmd []string, wd string, flags proc.LaunchFlags, debugInfoDirs []string, tty string, stdinPath string, stdoutOR proc.OutputRedirect, stderrOR proc.OutputRedirect) (*proc.TargetGroup, error) {
 	var (
 		process *exec.Cmd
 		err     error
@@ -69,7 +70,7 @@ func Launch(cmd []string, wd string, flags proc.LaunchFlags, debugInfoDirs []str
 
 	foreground := flags&proc.LaunchForeground != 0
 
-	stdin, stdout, stderr, closefn, err := openRedirects(redirects, foreground)
+	stdin, stdout, stderr, closefn, err := openRedirects(stdinPath, stdoutOR, stderrOR, foreground)
 	if err != nil {
 		return nil, err
 	}
@@ -83,7 +84,7 @@ func Launch(cmd []string, wd string, flags proc.LaunchFlags, debugInfoDirs []str
 	dbp := newProcess(0)
 	defer func() {
 		if err != nil && dbp.pid != 0 {
-			_ = dbp.Detach(true)
+			_ = detachWithoutGroup(dbp, true)
 		}
 	}()
 	dbp.execPtraceFunc(func() {
@@ -140,7 +141,15 @@ func Launch(cmd []string, wd string, flags proc.LaunchFlags, debugInfoDirs []str
 // Attach to an existing process with the given PID. Once attached, if
 // the DWARF information cannot be found in the binary, Delve will look
 // for external debug files in the directories passed in.
-func Attach(pid int, debugInfoDirs []string) (*proc.Target, error) {
+func Attach(pid int, waitFor *proc.WaitFor, debugInfoDirs []string) (*proc.TargetGroup, error) {
+	if waitFor.Valid() {
+		var err error
+		pid, err = WaitFor(waitFor)
+		if err != nil {
+			return nil, err
+		}
+	}
+
 	dbp := newProcess(pid)
 
 	var err error
@@ -155,7 +164,7 @@ func Attach(pid int, debugInfoDirs []string) (*proc.Target, error) {
 
 	tgt, err := dbp.initialize(findExecutable("", dbp.pid), debugInfoDirs)
 	if err != nil {
-		_ = dbp.Detach(false)
+		_ = detachWithoutGroup(dbp, false)
 		return nil, err
 	}
 
@@ -168,32 +177,79 @@ func Attach(pid int, debugInfoDirs []string) (*proc.Target, error) {
 	return tgt, nil
 }
 
-func initialize(dbp *nativeProcess) error {
-	comm, err := ioutil.ReadFile(fmt.Sprintf("/proc/%d/comm", dbp.pid))
+func isProcDir(name string) bool {
+	for _, ch := range name {
+		if ch < '0' || ch > '9' {
+			return false
+		}
+	}
+	return true
+}
+
+func waitForSearchProcess(pfx string, seen map[int]struct{}) (int, error) {
+	log := logflags.DebuggerLogger()
+	des, err := os.ReadDir("/proc")
+	if err != nil {
+		log.Errorf("error reading proc: %v", err)
+		return 0, nil
+	}
+	for _, de := range des {
+		if !de.IsDir() {
+			continue
+		}
+		name := de.Name()
+		if !isProcDir(name) {
+			continue
+		}
+		pid, _ := strconv.Atoi(name)
+		if _, isseen := seen[pid]; isseen {
+			continue
+		}
+		seen[pid] = struct{}{}
+		buf, err := os.ReadFile(filepath.Join("/proc", name, "cmdline"))
+		if err != nil {
+			// probably we just don't have permissions
+			continue
+		}
+		for i := range buf {
+			if buf[i] == 0 {
+				buf[i] = ' '
+			}
+		}
+		log.Debugf("waitfor: new process %q", string(buf))
+		if strings.HasPrefix(string(buf), pfx) {
+			return pid, nil
+		}
+	}
+	return 0, nil
+}
+
+func initialize(dbp *nativeProcess) (string, error) {
+	comm, err := os.ReadFile(fmt.Sprintf("/proc/%d/comm", dbp.pid))
 	if err == nil {
 		// removes newline character
 		comm = bytes.TrimSuffix(comm, []byte("\n"))
 	}
 
 	if comm == nil || len(comm) <= 0 {
-		stat, err := ioutil.ReadFile(fmt.Sprintf("/proc/%d/stat", dbp.pid))
+		stat, err := os.ReadFile(fmt.Sprintf("/proc/%d/stat", dbp.pid))
 		if err != nil {
-			return fmt.Errorf("could not read proc stat: %v", err)
+			return "", fmt.Errorf("could not read proc stat: %v", err)
 		}
 		expr := fmt.Sprintf("%d\\s*\\((.*)\\)", dbp.pid)
 		rexp, err := regexp.Compile(expr)
 		if err != nil {
-			return fmt.Errorf("regexp compile error: %v", err)
+			return "", fmt.Errorf("regexp compile error: %v", err)
 		}
 		match := rexp.FindSubmatch(stat)
 		if match == nil {
-			return fmt.Errorf("no match found using regexp '%s' in /proc/%d/stat", expr, dbp.pid)
+			return "", fmt.Errorf("no match found using regexp '%s' in /proc/%d/stat", expr, dbp.pid)
 		}
 		comm = match[1]
 	}
 	dbp.os.comm = strings.ReplaceAll(string(comm), "%", "%%")
 
-	return nil
+	return getCmdLine(dbp.pid), nil
 }
 
 func (dbp *nativeProcess) GetBufferedTracepoints() []ebpf.RawUProbeParams {
@@ -204,7 +260,7 @@ func (dbp *nativeProcess) GetBufferedTracepoints() []ebpf.RawUProbeParams {
 }
 
 // kill kills the target process.
-func (dbp *nativeProcess) kill() error {
+func (procgrp *processGroup) kill(dbp *nativeProcess) error {
 	if dbp.exited {
 		return nil
 	}
@@ -236,6 +292,11 @@ func (dbp *nativeProcess) requestManualStop() (err error) {
 	return sys.Kill(dbp.pid, sys.SIGTRAP)
 }
 
+const (
+	ptraceOptionsNormal     = syscall.PTRACE_O_TRACECLONE
+	ptraceOptionsFollowExec = syscall.PTRACE_O_TRACECLONE | syscall.PTRACE_O_TRACEVFORK | syscall.PTRACE_O_TRACEEXEC
+)
+
 // Attach to a newly created thread, and store that thread in our list of
 // known threads.
 func (dbp *nativeProcess) addThread(tid int, attach bool) (*nativeThread, error) {
@@ -243,6 +304,11 @@ func (dbp *nativeProcess) addThread(tid int, attach bool) (*nativeThread, error)
 		return thread, nil
 	}
 
+	ptraceOptions := ptraceOptionsNormal
+	if dbp.followExec {
+		ptraceOptions = ptraceOptionsFollowExec
+	}
+
 	var err error
 	if attach {
 		dbp.execPtraceFunc(func() { err = sys.PtraceAttach(tid) })
@@ -262,12 +328,12 @@ func (dbp *nativeProcess) addThread(tid int, attach bool) (*nativeThread, error)
 		}
 	}
 
-	dbp.execPtraceFunc(func() { err = syscall.PtraceSetOptions(tid, syscall.PTRACE_O_TRACECLONE) })
+	dbp.execPtraceFunc(func() { err = syscall.PtraceSetOptions(tid, ptraceOptions) })
 	if err == syscall.ESRCH {
 		if _, _, err = dbp.waitFast(tid); err != nil {
 			return nil, fmt.Errorf("error while waiting after adding thread: %d %s", tid, err)
 		}
-		dbp.execPtraceFunc(func() { err = syscall.PtraceSetOptions(tid, syscall.PTRACE_O_TRACECLONE) })
+		dbp.execPtraceFunc(func() { err = syscall.PtraceSetOptions(tid, ptraceOptions) })
 		if err == syscall.ESRCH {
 			return nil, err
 		}
@@ -317,8 +383,8 @@ func findExecutable(path string, pid int) string {
 	return path
 }
 
-func (dbp *nativeProcess) trapWait(pid int) (*nativeThread, error) {
-	return dbp.trapWaitInternal(pid, 0)
+func trapWait(procgrp *processGroup, pid int) (*nativeThread, error) {
+	return trapWaitInternal(procgrp, pid, 0)
 }
 
 type trapWaitOptions uint8
@@ -329,14 +395,21 @@ const (
 	trapWaitDontCallExitGuard
 )
 
-func (dbp *nativeProcess) trapWaitInternal(pid int, options trapWaitOptions) (*nativeThread, error) {
+func trapWaitInternal(procgrp *processGroup, pid int, options trapWaitOptions) (*nativeThread, error) {
+	var waitdbp *nativeProcess = nil
+	if len(procgrp.procs) == 1 {
+		// Note that waitdbp is only used to call (*nativeProcess).wait which will
+		// behave correctly if waitdbp == nil.
+		waitdbp = procgrp.procs[0]
+	}
+
 	halt := options&trapWaitHalt != 0
 	for {
 		wopt := 0
 		if options&trapWaitNohang != 0 {
 			wopt = sys.WNOHANG
 		}
-		wpid, status, err := dbp.wait(pid, wopt)
+		wpid, status, err := waitdbp.wait(pid, wopt)
 		if err != nil {
 			return nil, fmt.Errorf("wait err %s %d", err, pid)
 		}
@@ -346,14 +419,27 @@ func (dbp *nativeProcess) trapWaitInternal(pid int, options trapWaitOptions) (*n
 			}
 			continue
 		}
-		th, ok := dbp.threads[wpid]
-		if ok {
-			th.Status = (*waitStatus)(status)
+		dbp := procgrp.procForThread(wpid)
+		var th *nativeThread
+		if dbp != nil {
+			var ok bool
+			th, ok = dbp.threads[wpid]
+			if ok {
+				th.Status = (*waitStatus)(status)
+			}
+		} else {
+			dbp = procgrp.procs[0]
 		}
 		if status.Exited() {
 			if wpid == dbp.pid {
 				dbp.postExit()
-				return nil, proc.ErrProcessExited{Pid: wpid, Status: status.ExitStatus()}
+				if procgrp.numValid() == 0 {
+					return nil, proc.ErrProcessExited{Pid: wpid, Status: status.ExitStatus()}
+				}
+				if halt {
+					return nil, nil
+				}
+				continue
 			}
 			delete(dbp.threads, wpid)
 			continue
@@ -362,15 +448,24 @@ func (dbp *nativeProcess) trapWaitInternal(pid int, options trapWaitOptions) (*n
 			// Signaled means the thread was terminated due to a signal.
 			if wpid == dbp.pid {
 				dbp.postExit()
-				return nil, proc.ErrProcessExited{Pid: wpid, Status: -int(status.Signal())}
+				if procgrp.numValid() == 0 {
+					return nil, proc.ErrProcessExited{Pid: wpid, Status: -int(status.Signal())}
+				}
+				if halt {
+					return nil, nil
+				}
+				continue
 			}
 			// does this ever happen?
 			delete(dbp.threads, wpid)
 			continue
 		}
-		if status.StopSignal() == sys.SIGTRAP && status.TrapCause() == sys.PTRACE_EVENT_CLONE {
+		if status.StopSignal() == sys.SIGTRAP && (status.TrapCause() == sys.PTRACE_EVENT_CLONE || status.TrapCause() == sys.PTRACE_EVENT_VFORK) {
 			// A traced thread has cloned a new thread, grab the pid and
 			// add it to our list of traced threads.
+			// If TrapCause() is sys.PTRACE_EVENT_VFORK this is actually a new
+			// process, but treat it as a normal thread until exec happens, so that
+			// we can initialize the new process normally.
 			var cloned uint
 			dbp.execPtraceFunc(func() { cloned, err = sys.PtraceGetEventMsg(wpid) })
 			if err != nil {
@@ -394,7 +489,7 @@ func (dbp *nativeProcess) trapWaitInternal(pid int, options trapWaitOptions) (*n
 				dbp.threads[int(wpid)].os.running = false
 				return nil, nil
 			}
-			if err = th.Continue(); err != nil {
+			if err = th.resume(); err != nil {
 				if err == sys.ESRCH {
 					// thread died while we were adding it
 					delete(dbp.threads, th.ID)
@@ -402,13 +497,44 @@ func (dbp *nativeProcess) trapWaitInternal(pid int, options trapWaitOptions) (*n
 				}
 				return nil, fmt.Errorf("could not continue new thread %d %s", cloned, err)
 			}
-			if err = dbp.threads[int(wpid)].Continue(); err != nil {
+			if err = dbp.threads[int(wpid)].resume(); err != nil {
 				if err != sys.ESRCH {
 					return nil, fmt.Errorf("could not continue existing thread %d %s", wpid, err)
 				}
 			}
 			continue
 		}
+		if status.StopSignal() == sys.SIGTRAP && (status.TrapCause() == sys.PTRACE_EVENT_EXEC) {
+			// A thread called exec and we now have a new process. Retrieve the
+			// thread ID of the exec'ing thread with PtraceGetEventMsg to remove it
+			// and create a new nativeProcess object to track the new process.
+			var tid uint
+			dbp.execPtraceFunc(func() { tid, err = sys.PtraceGetEventMsg(wpid) })
+			if err == nil {
+				delete(dbp.threads, int(tid))
+			}
+			dbp = newChildProcess(procgrp.procs[0], wpid)
+			dbp.followExec = true
+			cmdline, _ := dbp.initializeBasic()
+			tgt, err := procgrp.add(dbp, dbp.pid, dbp.memthread, findExecutable("", dbp.pid), proc.StopLaunched, cmdline)
+			if err != nil {
+				return nil, err
+			}
+			if halt {
+				return nil, nil
+			}
+			if tgt != nil {
+				// If tgt is nil we decided we are not interested in debugging this
+				// process, and we have already detached from it.
+				err = dbp.threads[dbp.pid].resume()
+				if err != nil {
+					return nil, err
+				}
+			}
+			//TODO(aarzilli): if we want to give users the ability to stop the target
+			//group on exec here is where we should return
+			continue
+		}
 		if th == nil {
 			// Sometimes we get an unknown thread, ignore it?
 			continue
@@ -438,7 +564,10 @@ func (dbp *nativeProcess) trapWaitInternal(pid int, options trapWaitOptions) (*n
 			// do the same thing we do if a thread quit
 			if wpid == dbp.pid {
 				dbp.postExit()
-				return nil, proc.ErrProcessExited{Pid: wpid, Status: status.ExitStatus()}
+				if procgrp.numValid() == 0 {
+					return nil, proc.ErrProcessExited{Pid: wpid, Status: status.ExitStatus()}
+				}
+				continue
 			}
 			delete(dbp.threads, wpid)
 		}
@@ -458,7 +587,7 @@ func status(pid int, comm string) rune {
 		state rune
 	)
 
-	// The second field of /proc/pid/stat is the name of the task in parenthesis.
+	// The second field of /proc/pid/stat is the name of the task in parentheses.
 	// The name of the task is the base name of the executable for this process limited to TASK_COMM_LEN characters
 	// Since both parenthesis and spaces can appear inside the name of the task and no escaping happens we need to read the name of the executable first
 	// See: include/linux/sched.c:315 and include/linux/sched.c:1510
@@ -475,7 +604,7 @@ func (dbp *nativeProcess) waitFast(pid int) (int, *sys.WaitStatus, error) {
 
 func (dbp *nativeProcess) wait(pid, options int) (int, *sys.WaitStatus, error) {
 	var s sys.WaitStatus
-	if (pid != dbp.pid) || (options != 0) {
+	if (dbp == nil) || (pid != dbp.pid) || (options != 0) {
 		wpid, err := sys.Wait4(pid, &s, sys.WALL|options, nil)
 		return wpid, &s, err
 	}
@@ -505,53 +634,67 @@ func (dbp *nativeProcess) wait(pid, options int) (int, *sys.WaitStatus, error) {
 	}
 }
 
-func (dbp *nativeProcess) exitGuard(err error) error {
+func exitGuard(dbp *nativeProcess, procgrp *processGroup, err error) error {
 	if err != sys.ESRCH {
 		return err
 	}
 	if status(dbp.pid, dbp.os.comm) == statusZombie {
-		_, err := dbp.trapWaitInternal(-1, trapWaitDontCallExitGuard)
+		_, err := trapWaitInternal(procgrp, -1, trapWaitDontCallExitGuard)
 		return err
 	}
 
 	return err
 }
 
-func (dbp *nativeProcess) resume() error {
+func (procgrp *processGroup) resume() error {
 	// all threads stopped over a breakpoint are made to step over it
-	for _, thread := range dbp.threads {
-		if thread.CurrentBreakpoint.Breakpoint != nil {
-			if err := thread.StepInstruction(); err != nil {
-				return err
+	for _, dbp := range procgrp.procs {
+		if valid, _ := dbp.Valid(); valid {
+			for _, thread := range dbp.threads {
+				if thread.CurrentBreakpoint.Breakpoint != nil {
+					if err := procgrp.stepInstruction(thread); err != nil {
+						return err
+					}
+					thread.CurrentBreakpoint.Clear()
+				}
 			}
-			thread.CurrentBreakpoint.Clear()
 		}
 	}
 	// everything is resumed
-	for _, thread := range dbp.threads {
-		if err := thread.resume(); err != nil && err != sys.ESRCH {
-			return err
+	for _, dbp := range procgrp.procs {
+		if valid, _ := dbp.Valid(); valid {
+			for _, thread := range dbp.threads {
+				if err := thread.resume(); err != nil && err != sys.ESRCH {
+					return err
+				}
+			}
 		}
 	}
 	return nil
 }
 
 // stop stops all running threads and sets breakpoints
-func (dbp *nativeProcess) stop(trapthread *nativeThread) (*nativeThread, error) {
-	if dbp.exited {
-		return nil, proc.ErrProcessExited{Pid: dbp.Pid()}
+func (procgrp *processGroup) stop(cctx *proc.ContinueOnceContext, trapthread *nativeThread) (*nativeThread, error) {
+	if procgrp.numValid() == 0 {
+		return nil, proc.ErrProcessExited{Pid: procgrp.procs[0].pid}
 	}
 
-	for _, th := range dbp.threads {
-		th.os.setbp = false
+	for _, dbp := range procgrp.procs {
+		if dbp.exited {
+			continue
+		}
+		for _, th := range dbp.threads {
+			th.os.setbp = false
+		}
 	}
 	trapthread.os.setbp = true
 
 	// check if any other thread simultaneously received a SIGTRAP
 	for {
-		th, err := dbp.trapWaitInternal(-1, trapWaitNohang)
+		th, err := trapWaitInternal(procgrp, -1, trapWaitNohang)
 		if err != nil {
-			return nil, dbp.exitGuard(err)
+			p := procgrp.procForThread(th.ID)
+			return nil, exitGuard(p, procgrp, err)
 		}
 		if th == nil {
 			break
@@ -559,10 +702,20 @@ func (dbp *nativeProcess) stop(trapthread *nativeThread) (*nativeThread, error)
 	}
 
 	// stop all threads that are still running
-	for _, th := range dbp.threads {
-		if th.os.running {
-			if err := th.stop(); err != nil {
-				return nil, dbp.exitGuard(err)
+	for _, dbp := range procgrp.procs {
+		if dbp.exited {
+			continue
+		}
+		for _, th := range dbp.threads {
+			if th.os.running {
+				if err := th.stop(); err != nil {
+					if err == sys.ESRCH {
+						// thread exited
+						delete(dbp.threads, th.ID)
+					} else {
+						return nil, exitGuard(dbp, procgrp, err)
+					}
+				}
 			}
 		}
 	}
@@ -570,26 +723,60 @@ func (dbp *nativeProcess) stop(trapthread *nativeThread) (*nativeThread, error)
 	// wait for all threads to stop
 	for {
 		allstopped := true
-		for _, th := range dbp.threads {
-			if th.os.running {
-				allstopped = false
-				break
+		for _, dbp := range procgrp.procs {
+			if dbp.exited {
+				continue
+			}
+			for _, th := range dbp.threads {
+				if th.os.running {
+					allstopped = false
+					break
+				}
 			}
 		}
 		if allstopped {
 			break
 		}
-		_, err := dbp.trapWaitInternal(-1, trapWaitHalt)
+		_, err := trapWaitInternal(procgrp, -1, trapWaitHalt)
 		if err != nil {
 			return nil, err
 		}
 	}
 
-	if err := linutil.ElfUpdateSharedObjects(dbp); err != nil {
-		return nil, err
+	switchTrapthread := false
+
+	for _, dbp := range procgrp.procs {
+		if dbp.exited {
+			continue
+		}
+		err := stop1(cctx, dbp, trapthread, &switchTrapthread)
+		if err != nil {
+			return nil, err
+		}
 	}
 
-	switchTrapthread := false
+	if switchTrapthread {
+		trapthreadID := trapthread.ID
+		trapthread = nil
+		for _, dbp := range procgrp.procs {
+			if dbp.exited {
+				continue
+			}
+			for _, th := range dbp.threads {
+				if th.os.setbp && th.ThreadID() != trapthreadID {
+					return th, nil
+				}
+			}
+		}
+	}
+
+	return trapthread, nil
+}
+
+func stop1(cctx *proc.ContinueOnceContext, dbp *nativeProcess, trapthread *nativeThread, switchTrapthread *bool) error {
+	if err := linutil.ElfUpdateSharedObjects(dbp); err != nil {
+		return err
+	}
 
 	// set breakpoints on SIGTRAP threads
 	var err1 error
@@ -617,9 +804,7 @@ func (dbp *nativeProcess) stop(trapthread *nativeThread) (*nativeThread, error)
 		if th.CurrentBreakpoint.Breakpoint == nil && th.os.setbp && (th.Status != nil) && ((*sys.WaitStatus)(th.Status).StopSignal() == sys.SIGTRAP) && dbp.BinInfo().Arch.BreakInstrMovesPC() {
 			manualStop := false
 			if th.ThreadID() == trapthread.ThreadID() {
-				dbp.stopMu.Lock()
-				manualStop = dbp.manualStopRequested
-				dbp.stopMu.Unlock()
+				manualStop = cctx.GetManualStopRequested()
 			}
 			if !manualStop && th.os.phantomBreakpointPC == pc {
 				// Thread received a SIGTRAP but we don't have a breakpoint for it and
@@ -650,27 +835,13 @@ func (dbp *nativeProcess) stop(trapthread *nativeThread) (*nativeThread, error)
 						// Will switch to a different thread for trapthread because we don't
 						// want pkg/proc to believe that this thread was stopped by a
 						// hardcoded breakpoint.
-						switchTrapthread = true
+						*switchTrapthread = true
 					}
 				}
 			}
 		}
 	}
-	if err1 != nil {
-		return nil, err1
-	}
-
-	if switchTrapthread {
-		trapthreadID := trapthread.ID
-		trapthread = nil
-		for _, th := range dbp.threads {
-			if th.os.setbp && th.ThreadID() != trapthreadID {
-				return th, nil
-			}
-		}
-	}
-
-	return trapthread, nil
+	return err1
 }
 
 func (dbp *nativeProcess) detach(kill bool) error {
@@ -697,7 +868,7 @@ func (dbp *nativeProcess) detach(kill bool) error {
 // EntryPoint will return the process entry point address, useful for
 // debugging PIEs.
 func (dbp *nativeProcess) EntryPoint() (uint64, error) {
-	auxvbuf, err := ioutil.ReadFile(fmt.Sprintf("/proc/%d/auxv", dbp.pid))
+	auxvbuf, err := os.ReadFile(fmt.Sprintf("/proc/%d/auxv", dbp.pid))
 	if err != nil {
 		return 0, fmt.Errorf("could not read auxiliary vector: %v", err)
 	}
@@ -705,36 +876,124 @@ func (dbp *nativeProcess) EntryPoint() (uint64, error) {
 	return linutil.EntryPointFromAuxv(auxvbuf, dbp.bi.Arch.PtrSize()), nil
 }
 
-func (dbp *nativeProcess) SetUProbe(fnName string, args []ebpf.UProbeArgMap) error {
+func (dbp *nativeProcess) SetUProbe(fnName string, goidOffset int64, args []ebpf.UProbeArgMap) error {
 	// Lazily load and initialize the BPF program upon request to set a uprobe.
 	if dbp.os.ebpf == nil {
-		dbp.os.ebpf, _ = ebpf.LoadEBPFTracingProgram()
+		var err error
+		dbp.os.ebpf, err = ebpf.LoadEBPFTracingProgram(dbp.bi.Images[0].Path)
+		if err != nil {
+			return err
+		}
 	}
 
-	// We only allow up to 6 args for a BPF probe.
+	// We only allow up to 12 args for a BPF probe.
+	// 6 inputs + 6 outputs.
 	// Return early if we have more.
-	if len(args) > 6 {
-		return errors.New("too many arguments in traced function, max is 6")
+	if len(args) > 12 {
+		return errors.New("too many arguments in traced function, max is 12 input+return")
+	}
+
+	fns := dbp.bi.LookupFunc()[fnName]
+	if len(fns) != 1 {
+		return fmt.Errorf("could not find function: %s", fnName)
+	}
+	fn := fns[0]
+
+	offset, err := dbp.BinInfo().GStructOffset(dbp.Memory())
+	if err != nil {
+		return err
+	}
+	key := fn.Entry
+	err = dbp.os.ebpf.UpdateArgMap(key, goidOffset, args, offset, false)
+	if err != nil {
+		return err
 	}
 
 	debugname := dbp.bi.Images[0].Path
-	offset, err := ebpf.SymbolToOffset(debugname, fnName)
+
+	// First attach a uprobe at all return addresses. We do this instead of using a uretprobe
+	// for two reasons:
+	// 1. uretprobes do not play well with Go
+	// 2. uretprobes seem to not restore the function return addr on the stack when removed, destroying any
+	//    kind of workaround we could come up with.
+	// TODO(derekparker): this whole thing could likely be optimized a bit.
+	img := dbp.BinInfo().PCToImage(fn.Entry)
+	f, err := elf.Open(img.Path)
 	if err != nil {
-		return err
-	}
-	err = dbp.os.ebpf.AttachUprobe(dbp.Pid(), debugname, offset)
-	if err != nil {
-		return err
-	}
-	fn, ok := dbp.bi.LookupFunc[fnName]
-	if !ok {
-		return fmt.Errorf("could not find function: %s", fnName)
+		return fmt.Errorf("could not open elf file to resolve symbol offset: %w", err)
 	}
 
-	key := fn.Entry
-	return dbp.os.ebpf.UpdateArgMap(key, args)
+	var regs proc.Registers
+	mem := dbp.Memory()
+	regs, _ = dbp.memthread.Registers()
+	instructions, err := proc.Disassemble(mem, regs, &proc.BreakpointMap{}, dbp.BinInfo(), fn.Entry, fn.End)
+	if err != nil {
+		return err
+	}
+
+	var addrs []uint64
+	for _, instruction := range instructions {
+		if instruction.IsRet() {
+			addrs = append(addrs, instruction.Loc.PC)
+		}
+	}
+	addrs = append(addrs, proc.FindDeferReturnCalls(instructions)...)
+	for _, addr := range addrs {
+		err := dbp.os.ebpf.UpdateArgMap(addr, goidOffset, args, offset, true)
+		if err != nil {
+			return err
+		}
+		off, err := ebpf.AddressToOffset(f, addr)
+		if err != nil {
+			return err
+		}
+		err = dbp.os.ebpf.AttachUprobe(dbp.pid, debugname, off)
+		if err != nil {
+			return err
+		}
+	}
+
+	off, err := ebpf.AddressToOffset(f, fn.Entry)
+	if err != nil {
+		return err
+	}
+
+	return dbp.os.ebpf.AttachUprobe(dbp.pid, debugname, off)
+}
+
+// FollowExec enables (or disables) follow exec mode
+func (dbp *nativeProcess) FollowExec(v bool) error {
+	dbp.followExec = v
+	ptraceOptions := ptraceOptionsNormal
+	if dbp.followExec {
+		ptraceOptions = ptraceOptionsFollowExec
+	}
+	var err error
+	dbp.execPtraceFunc(func() {
+		for tid := range dbp.threads {
+			err = syscall.PtraceSetOptions(tid, ptraceOptions)
+			if err != nil {
+				return
+			}
+		}
+	})
+	return err
 }
 
 func killProcess(pid int) error {
 	return sys.Kill(pid, sys.SIGINT)
 }
+
+func getCmdLine(pid int) string {
+	buf, _ := os.ReadFile(fmt.Sprintf("/proc/%d/cmdline", pid))
+	args := strings.SplitN(string(buf), "\x00", -1)
+	for i := range args {
+		if strings.Contains(args[i], " ") {
+			args[i] = strconv.Quote(args[i])
+		}
+	}
+	if len(args) > 0 && args[len(args)-1] == "" {
+		args = args[:len(args)-1]
+	}
+	return strings.Join(args, " ")
+}
