@@ -3,7 +3,6 @@ package service_test
 import (
 	"flag"
 	"fmt"
-	"io/ioutil"
 	"math/rand"
 	"net"
 	"net/rpc"
@@ -23,6 +22,7 @@ import (
 
 	"github.com/go-delve/delve/pkg/goversion"
 	"github.com/go-delve/delve/pkg/logflags"
+	"github.com/go-delve/delve/pkg/proc"
 	"github.com/go-delve/delve/service"
 	"github.com/go-delve/delve/service/api"
 	"github.com/go-delve/delve/service/rpc2"
@@ -55,12 +55,12 @@ func TestMain(m *testing.M) {
 }
 
 func withTestClient2(name string, t *testing.T, fn func(c service.Client)) {
-	withTestClient2Extended(name, t, 0, [3]string{}, func(c service.Client, fixture protest.Fixture) {
+	withTestClient2Extended(name, t, 0, [3]string{}, nil, func(c service.Client, fixture protest.Fixture) {
 		fn(c)
 	})
 }
 
-func startServer(name string, buildFlags protest.BuildFlags, t *testing.T, redirects [3]string) (clientConn net.Conn, fixture protest.Fixture) {
+func startServer(name string, buildFlags protest.BuildFlags, t *testing.T, redirects [3]string, args []string) (clientConn net.Conn, fixture protest.Fixture) {
 	if testBackend == "rr" {
 		protest.MustHaveRecordingAllowed(t)
 	}
@@ -77,14 +77,16 @@ func startServer(name string, buildFlags protest.BuildFlags, t *testing.T, redir
 	}
 	server := rpccommon.NewServer(&service.Config{
 		Listener:    listener,
-		ProcessArgs: []string{fixture.Path},
+		ProcessArgs: append([]string{fixture.Path}, args...),
 		Debugger: debugger.Config{
 			Backend:        testBackend,
 			CheckGoVersion: true,
 			Packages:       []string{fixture.Source},
 			BuildFlags:     "", // build flags can be an empty string here because the only test that uses it, does not set special flags.
 			ExecuteKind:    debugger.ExecutingGeneratedFile,
-			Redirects:      redirects,
+			Stdin:          redirects[0],
+			Stdout:         proc.OutputRedirect{Path: redirects[1]},
+			Stderr:         proc.OutputRedirect{Path: redirects[2]},
 		},
 	})
 	if err := server.Run(); err != nil {
@@ -93,8 +95,8 @@ func startServer(name string, buildFlags protest.BuildFlags, t *testing.T, redir
 	return clientConn, fixture
 }
 
-func withTestClient2Extended(name string, t *testing.T, buildFlags protest.BuildFlags, redirects [3]string, fn func(c service.Client, fixture protest.Fixture)) {
-	clientConn, fixture := startServer(name, buildFlags, t, redirects)
+func withTestClient2Extended(name string, t *testing.T, buildFlags protest.BuildFlags, redirects [3]string, args []string, fn func(c service.Client, fixture protest.Fixture)) {
+	clientConn, fixture := startServer(name, buildFlags, t, redirects, args)
 	client := rpc2.NewClientFromConn(clientConn)
 	defer func() {
 		client.Detach(true)
@@ -228,9 +230,9 @@ func main() {
 
 func TestRestart_rebuild(t *testing.T) {
 	// In the original fixture file the env var tested for is SOMEVAR.
-	os.Setenv("SOMEVAR", "bah")
+	t.Setenv("SOMEVAR", "bah")
 
-	withTestClient2Extended("testenv", t, 0, [3]string{}, func(c service.Client, f protest.Fixture) {
+	withTestClient2Extended("testenv", t, 0, [3]string{}, nil, func(c service.Client, f protest.Fixture) {
 		<-c.Continue()
 
 		var1, err := c.EvalVariable(api.EvalScope{GoroutineID: -1}, "x", normalLoadConfig)
@@ -243,19 +245,19 @@ func TestRestart_rebuild(t *testing.T) {
 		fi, err := os.Stat(f.Source)
 		assertNoError(err, t, "Stat fixture.Source")
 
-		originalSource, err := ioutil.ReadFile(f.Source)
+		originalSource, err := os.ReadFile(f.Source)
 		assertNoError(err, t, "Reading original source")
 
 		// Ensure we write the original source code back after the test exits.
-		defer ioutil.WriteFile(f.Source, originalSource, fi.Mode())
+		defer os.WriteFile(f.Source, originalSource, fi.Mode())
 
 		// Write modified source code to the fixture file.
-		err = ioutil.WriteFile(f.Source, []byte(modifiedSource), fi.Mode())
+		err = os.WriteFile(f.Source, []byte(modifiedSource), fi.Mode())
 		assertNoError(err, t, "Writing modified source")
 
 		// First set our new env var and ensure later that the
 		// modified source code picks it up.
-		os.Setenv("SOMEMODIFIEDVAR", "foobar")
+		t.Setenv("SOMEMODIFIEDVAR", "foobar")
 
 		// Restart the program, rebuilding from source.
 		_, err = c.Restart(true)
@@ -610,7 +612,7 @@ func TestClientServer_toggleAmendedBreakpoint(t *testing.T) {
 			t.Fatal(err)
 		}
 		if amended.Cond == "" {
-			t.Fatal("breakpoint amendedments not preserved after toggle")
+			t.Fatal("breakpoint amendments not preserved after toggle")
 		}
 	})
 }
@@ -862,7 +864,7 @@ func TestClientServer_traceContinue(t *testing.T) {
 					t.Fatalf("No goroutine information")
 				}
 
-				if len(bpi.Stacktrace) <= 0 {
+				if len(bpi.Stacktrace) == 0 {
 					t.Fatalf("No stacktrace\n")
 				}
 
@@ -943,6 +945,9 @@ func TestClientServer_traceContinue2(t *testing.T) {
 }
 
 func TestClientServer_FindLocations(t *testing.T) {
+	if runtime.GOARCH == "ppc64le" && buildMode == "pie" {
+		t.Skip("pie mode broken on ppc64le")
+	}
 	withTestClient2("locationsprog", t, func(c service.Client) {
 		someFunctionCallAddr := findLocationHelper(t, c, "locationsprog.go:26", false, 1, 0)[0]
 		someFunctionLine1 := findLocationHelper(t, c, "locationsprog.go:27", false, 1, 0)[0]
@@ -991,14 +996,14 @@ func TestClientServer_FindLocations(t *testing.T) {
 
 		findLocationHelper(t, c, `*amap["k"]`, false, 1, findLocationHelper(t, c, `amap["k"]`, false, 1, 0)[0])
 
-		locsNoSubst, _ := c.FindLocation(api.EvalScope{GoroutineID: -1}, "_fixtures/locationsprog.go:35", false, nil)
+		locsNoSubst, _, _ := c.FindLocation(api.EvalScope{GoroutineID: -1}, "_fixtures/locationsprog.go:35", false, nil)
 		sep := "/"
 		if strings.Contains(locsNoSubst[0].File, "\\") {
 			sep = "\\"
 		}
-		substRules := [][2]string{[2]string{strings.Replace(locsNoSubst[0].File, "locationsprog.go", "", 1), strings.Replace(locsNoSubst[0].File, "_fixtures"+sep+"locationsprog.go", "nonexistent", 1)}}
+		substRules := [][2]string{{strings.Replace(locsNoSubst[0].File, "locationsprog.go", "", 1), strings.Replace(locsNoSubst[0].File, "_fixtures"+sep+"locationsprog.go", "nonexistent", 1)}}
 		t.Logf("substitute rules: %q -> %q", substRules[0][0], substRules[0][1])
-		locsSubst, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, "nonexistent/locationsprog.go:35", false, substRules)
+		locsSubst, _, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, "nonexistent/locationsprog.go:35", false, substRules)
 		if err != nil {
 			t.Fatalf("FindLocation(locationsprog.go:35) with substitute rules: %v", err)
 		}
@@ -1019,7 +1024,7 @@ func TestClientServer_FindLocations(t *testing.T) {
 		findLocationHelper(t, c, "main.stacktraceme", false, 1, stacktracemeAddr)
 	})
 
-	withTestClient2Extended("locationsUpperCase", t, 0, [3]string{}, func(c service.Client, fixture protest.Fixture) {
+	withTestClient2Extended("locationsUpperCase", t, 0, [3]string{}, nil, func(c service.Client, fixture protest.Fixture) {
 		// Upper case
 		findLocationHelper(t, c, "locationsUpperCase.go:6", false, 1, 0)
 
@@ -1108,7 +1113,7 @@ func TestClientServer_FindLocations(t *testing.T) {
 }
 
 func findLocationHelper2(t *testing.T, c service.Client, loc string, checkLoc *api.Location) *api.Location {
-	locs, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, loc, false, nil)
+	locs, _, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, loc, false, nil)
 	if err != nil {
 		t.Fatalf("FindLocation(%q) -> error %v", loc, err)
 	}
@@ -1207,6 +1212,9 @@ func TestClientServer_FullStacktrace(t *testing.T) {
 	if runtime.GOOS == "darwin" && runtime.GOARCH == "arm64" {
 		t.Skip("cgo doesn't work on darwin/arm64")
 	}
+	if runtime.GOARCH == "ppc64le" && buildMode == "pie" {
+		t.Skip("pie mode broken on ppc64le")
+	}
 
 	lenient := false
 	if runtime.GOOS == "windows" {
@@ -1305,7 +1313,7 @@ func TestIssue355(t *testing.T) {
 	// After the target process has terminated should return an error but not crash
 	protest.AllowRecording(t)
 	withTestClient2("continuetestprog", t, func(c service.Client) {
-		bp, err := c.CreateBreakpoint(&api.Breakpoint{FunctionName: "main.sayhi", Line: -1})
+		_, err := c.CreateBreakpoint(&api.Breakpoint{FunctionName: "main.sayhi", Line: -1})
 		assertNoError(err, t, "CreateBreakpoint()")
 		ch := c.Continue()
 		state := <-ch
@@ -1334,14 +1342,6 @@ func TestIssue355(t *testing.T) {
 		assertErrorOrExited(s, err, t, "SwitchGoroutine()")
 		s, err = c.Halt()
 		assertErrorOrExited(s, err, t, "Halt()")
-		_, err = c.CreateBreakpoint(&api.Breakpoint{FunctionName: "main.main", Line: -1})
-		if testBackend != "rr" {
-			assertError(err, t, "CreateBreakpoint()")
-		}
-		_, err = c.ClearBreakpoint(bp.ID)
-		if testBackend != "rr" {
-			assertError(err, t, "ClearBreakpoint()")
-		}
 		_, err = c.ListThreads()
 		assertError(err, t, "ListThreads()")
 		_, err = c.GetThread(tid)
@@ -1359,7 +1359,7 @@ func TestIssue355(t *testing.T) {
 		assertError(err, t, "ListGoroutines()")
 		_, err = c.Stacktrace(gid, 10, 0, &normalLoadConfig)
 		assertError(err, t, "Stacktrace()")
-		_, err = c.FindLocation(api.EvalScope{GoroutineID: gid}, "+1", false, nil)
+		_, _, err = c.FindLocation(api.EvalScope{GoroutineID: gid}, "+1", false, nil)
 		assertError(err, t, "FindLocation()")
 		_, err = c.DisassemblePC(api.EvalScope{GoroutineID: -1}, 0x40100, api.IntelFlavour)
 		assertError(err, t, "DisassemblePC()")
@@ -1367,7 +1367,10 @@ func TestIssue355(t *testing.T) {
 }
 
 func TestDisasm(t *testing.T) {
-	// Tests that disassembling by PC, range, and current PC all yeld similar results
+	if runtime.GOARCH == "ppc64le" {
+		t.Skip("skipped on ppc64le: broken")
+	}
+	// Tests that disassembling by PC, range, and current PC all yield similar results
 	// Tests that disassembly by current PC will return a disassembly containing the instruction at PC
 	// Tests that stepping on a calculated CALL instruction will yield a disassembly that contains the
 	// effective destination of the CALL instruction
@@ -1376,7 +1379,7 @@ func TestDisasm(t *testing.T) {
 		state := <-ch
 		assertNoError(state.Err, t, "Continue()")
 
-		locs, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, "main.main", false, nil)
+		locs, _, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, "main.main", false, nil)
 		assertNoError(err, t, "FindLocation()")
 		if len(locs) != 1 {
 			t.Fatalf("wrong number of locations for main.main: %d", len(locs))
@@ -1496,9 +1499,6 @@ func TestNegativeStackDepthBug(t *testing.T) {
 }
 
 func TestClientServer_CondBreakpoint(t *testing.T) {
-	if runtime.GOOS == "freebsd" {
-		t.Skip("test is not valid on FreeBSD")
-	}
 	protest.AllowRecording(t)
 	withTestClient2("parallel_next", t, func(c service.Client) {
 		bp, err := c.CreateBreakpoint(&api.Breakpoint{FunctionName: "main.sayhi", Line: 1})
@@ -1638,7 +1638,7 @@ func TestTypesCommand(t *testing.T) {
 func TestIssue406(t *testing.T) {
 	protest.AllowRecording(t)
 	withTestClient2("issue406", t, func(c service.Client) {
-		locs, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, "issue406.go:146", false, nil)
+		locs, _, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, "issue406.go:146", false, nil)
 		assertNoError(err, t, "FindLocation()")
 		_, err = c.CreateBreakpoint(&api.Breakpoint{Addr: locs[0].PC})
 		assertNoError(err, t, "CreateBreakpoint()")
@@ -1724,7 +1724,7 @@ func TestClientServer_FpRegisters(t *testing.T) {
 		{"XMM12", "…[ZMM12hh] 0x3ff66666666666663ff4cccccccccccd"},
 	}
 	protest.AllowRecording(t)
-	withTestClient2Extended("fputest/", t, 0, [3]string{}, func(c service.Client, fixture protest.Fixture) {
+	withTestClient2Extended("fputest/", t, 0, [3]string{}, nil, func(c service.Client, fixture protest.Fixture) {
 		_, err := c.CreateBreakpoint(&api.Breakpoint{File: filepath.Join(fixture.BuildDir, "fputest.go"), Line: 25})
 		assertNoError(err, t, "CreateBreakpoint")
 
@@ -1972,7 +1972,7 @@ func TestClientServerConsistentExit(t *testing.T) {
 
 		// Ensure future commands also return the correct exit status.
 		// Previously there was a bug where the command which prompted the
-		// process to exit (continue, next, etc...) would return the corrent
+		// process to exit (continue, next, etc...) would return the current
 		// exit status but subsequent commands would return an incorrect exit
 		// status of 0. To test this we simply repeat the 'next' command and
 		// ensure we get the correct response again.
@@ -2011,19 +2011,6 @@ func TestClientServer_StepOutReturn(t *testing.T) {
 		stridx := 0
 		numidx := 1
 
-		if !goversion.VersionAfterOrEqual(runtime.Version(), 1, 12) {
-			// in 1.11 and earlier the order of return values in DWARF is
-			// unspecified, in 1.11 and later it follows the order of definition
-			// specified by the user
-			for i := range ret {
-				if ret[i].Name == "str" {
-					stridx = i
-					numidx = 1 - i
-					break
-				}
-			}
-		}
-
 		if ret[stridx].Name != "str" {
 			t.Fatalf("(str) bad return value name %s", ret[stridx].Name)
 		}
@@ -2121,6 +2108,10 @@ func TestForceStopWhileContinue(t *testing.T) {
 }
 
 func TestClientServerFunctionCall(t *testing.T) {
+	if buildMode == "pie" && runtime.GOARCH == "ppc64le" {
+		t.Skip("Debug function call Test broken in PIE mode")
+	}
+
 	protest.MustSupportFunctionCalls(t, testBackend)
 	withTestClient2("fncall", t, func(c service.Client) {
 		c.SetReturnValuesLoadConfig(&normalLoadConfig)
@@ -2150,33 +2141,10 @@ func TestClientServerFunctionCall(t *testing.T) {
 	})
 }
 
-func TestClientServerFunctionCallBadPos(t *testing.T) {
-	protest.MustSupportFunctionCalls(t, testBackend)
-	if goversion.VersionAfterOrEqual(runtime.Version(), 1, 12) {
-		t.Skip("this is a safe point for Go 1.12")
-	}
-	withTestClient2("fncall", t, func(c service.Client) {
-		loc, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, "fmt/print.go:649", false, nil)
-		assertNoError(err, t, "could not find location")
-
-		_, err = c.CreateBreakpoint(&api.Breakpoint{File: loc[0].File, Line: loc[0].Line})
-		assertNoError(err, t, "CreateBreakpoin")
-
-		state := <-c.Continue()
-		assertNoError(state.Err, t, "Continue()")
-
-		state = <-c.Continue()
-		assertNoError(state.Err, t, "Continue()")
-
-		c.SetReturnValuesLoadConfig(&normalLoadConfig)
-		state, err = c.Call(-1, "main.call1(main.zero, main.zero)", false)
-		if err == nil || err.Error() != "call not at safe point" {
-			t.Fatalf("wrong error or no error: %v", err)
-		}
-	})
-}
-
 func TestClientServerFunctionCallPanic(t *testing.T) {
+	if buildMode == "pie" && runtime.GOARCH == "ppc64le" {
+		t.Skip("Debug function call Test broken in PIE mode")
+	}
 	protest.MustSupportFunctionCalls(t, testBackend)
 	withTestClient2("fncall", t, func(c service.Client) {
 		c.SetReturnValuesLoadConfig(&normalLoadConfig)
@@ -2232,12 +2200,10 @@ func TestAncestors(t *testing.T) {
 	if !goversion.VersionAfterOrEqual(runtime.Version(), 1, 11) {
 		t.Skip("not supported on Go <= 1.10")
 	}
-	savedGodebug := os.Getenv("GODEBUG")
-	os.Setenv("GODEBUG", "tracebackancestors=100")
-	defer os.Setenv("GODEBUG", savedGodebug)
+	t.Setenv("GODEBUG", "tracebackancestors=100")
 	withTestClient2("testnextprog", t, func(c service.Client) {
 		_, err := c.CreateBreakpoint(&api.Breakpoint{FunctionName: "main.testgoroutine", Line: -1})
-		assertNoError(err, t, "CreateBreakpoin")
+		assertNoError(err, t, "CreateBreakpoint")
 		state := <-c.Continue()
 		assertNoError(state.Err, t, "Continue()")
 		ancestors, err := c.Ancestors(-1, 1000, 1000)
@@ -2276,7 +2242,7 @@ func (c *brokenRPCClient) call(method string, args, reply interface{}) error {
 }
 
 func TestUnknownMethodCall(t *testing.T) {
-	clientConn, _ := startServer("continuetestprog", 0, t, [3]string{})
+	clientConn, _ := startServer("continuetestprog", 0, t, [3]string{}, nil)
 	client := &brokenRPCClient{jsonrpc.NewClient(clientConn)}
 	client.call("SetApiVersion", api.SetAPIVersionIn{APIVersion: 2}, &api.SetAPIVersionOut{})
 	defer client.Detach(true)
@@ -2291,7 +2257,7 @@ func TestUnknownMethodCall(t *testing.T) {
 func TestIssue1703(t *testing.T) {
 	// Calling Disassemble when there is no current goroutine should work.
 	withTestClient2("testnextprog", t, func(c service.Client) {
-		locs, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, "main.main", true, nil)
+		locs, _, err := c.FindLocation(api.EvalScope{GoroutineID: -1}, "main.main", true, nil)
 		assertNoError(err, t, "FindLocation")
 		t.Logf("FindLocation: %#v", locs)
 		text, err := c.DisassemblePC(api.EvalScope{GoroutineID: -1}, locs[0].PC, api.IntelFlavour)
@@ -2308,7 +2274,7 @@ func TestRerecord(t *testing.T) {
 	withTestClient2("testrerecord", t, func(c service.Client) {
 		fp := testProgPath(t, "testrerecord")
 		_, err := c.CreateBreakpoint(&api.Breakpoint{File: fp, Line: 10})
-		assertNoError(err, t, "CreateBreakpoin")
+		assertNoError(err, t, "CreateBreakpoint")
 
 		gett := func() int {
 			state := <-c.Continue()
@@ -2420,7 +2386,7 @@ func TestClearLogicalBreakpoint(t *testing.T) {
 	// Clearing a logical breakpoint should clear all associated physical
 	// breakpoints.
 	// Issue #1955.
-	withTestClient2Extended("testinline", t, protest.EnableInlining, [3]string{}, func(c service.Client, fixture protest.Fixture) {
+	withTestClient2Extended("testinline", t, protest.EnableInlining, [3]string{}, nil, func(c service.Client, fixture protest.Fixture) {
 		bp, err := c.CreateBreakpoint(&api.Breakpoint{FunctionName: "main.inlineThis"})
 		assertNoError(err, t, "CreateBreakpoint()")
 		t.Logf("breakpoint set at %#v", bp.Addrs)
@@ -2446,10 +2412,10 @@ func TestRedirects(t *testing.T) {
 		outfile = "redirect-output.txt"
 	)
 	protest.AllowRecording(t)
-	withTestClient2Extended("redirect", t, 0, [3]string{infile, outfile, ""}, func(c service.Client, fixture protest.Fixture) {
+	withTestClient2Extended("redirect", t, 0, [3]string{infile, outfile, ""}, nil, func(c service.Client, fixture protest.Fixture) {
 		outpath := filepath.Join(fixture.BuildDir, outfile)
 		<-c.Continue()
-		buf, err := ioutil.ReadFile(outpath)
+		buf, err := os.ReadFile(outpath)
 		assertNoError(err, t, "Reading output file")
 		t.Logf("output %q", buf)
 		if !strings.HasPrefix(string(buf), "Redirect test") {
@@ -2460,7 +2426,7 @@ func TestRedirects(t *testing.T) {
 			_, err = c.Restart(false)
 			assertNoError(err, t, "Restart")
 			<-c.Continue()
-			buf2, err := ioutil.ReadFile(outpath)
+			buf2, err := os.ReadFile(outpath)
 			t.Logf("output %q", buf2)
 			assertNoError(err, t, "Reading output file (second time)")
 			if !strings.HasPrefix(string(buf2), "Redirect test") {
@@ -2580,7 +2546,7 @@ func TestToggleBreakpointRestart(t *testing.T) {
 }
 
 func TestStopServerWithClosedListener(t *testing.T) {
-	// Checks that the error erturned by listener.Accept() is ignored when we
+	// Checks that the error returned by listener.Accept() is ignored when we
 	// are trying to shutdown. See issue #1633.
 	if testBackend == "rr" || buildMode == "pie" {
 		t.Skip("N/A")
@@ -2615,7 +2581,7 @@ func TestGoroutinesGrouping(t *testing.T) {
 	withTestClient2("goroutinegroup", t, func(c service.Client) {
 		state := <-c.Continue()
 		assertNoError(state.Err, t, "Continue")
-		_, ggrp, _, _, err := c.ListGoroutinesWithFilter(0, 0, nil, &api.GoroutineGroupingOptions{GroupBy: api.GoroutineLabel, GroupByKey: "name", MaxGroupMembers: 5, MaxGroups: 10})
+		_, ggrp, _, _, err := c.ListGoroutinesWithFilter(0, 0, nil, &api.GoroutineGroupingOptions{GroupBy: api.GoroutineLabel, GroupByKey: "name", MaxGroupMembers: 5, MaxGroups: 10}, nil)
 		assertNoError(err, t, "ListGoroutinesWithFilter (group by label)")
 		t.Logf("%#v\n", ggrp)
 		if len(ggrp) < 5 {
@@ -2628,7 +2594,7 @@ func TestGoroutinesGrouping(t *testing.T) {
 				break
 			}
 		}
-		gs, _, _, _, err := c.ListGoroutinesWithFilter(0, 0, []api.ListGoroutinesFilter{{Kind: api.GoroutineLabel, Arg: "name="}}, nil)
+		gs, _, _, _, err := c.ListGoroutinesWithFilter(0, 0, []api.ListGoroutinesFilter{{Kind: api.GoroutineLabel, Arg: "name="}}, nil, nil)
 		assertNoError(err, t, "ListGoroutinesWithFilter (filter unnamed)")
 		if len(gs) != unnamedCount {
 			t.Errorf("wrong number of goroutines returned by filter: %d (expected %d)\n", len(gs), unnamedCount)
@@ -2774,7 +2740,7 @@ func TestClientServer_SinglelineStringFormattedWithBigInts(t *testing.T) {
 	if runtime.GOARCH != "amd64" {
 		t.Skip("N/A")
 	}
-	withTestClient2Extended("xmm0print/", t, 0, [3]string{}, func(c service.Client, fixture protest.Fixture) {
+	withTestClient2Extended("xmm0print/", t, 0, [3]string{}, nil, func(c service.Client, fixture protest.Fixture) {
 		_, err := c.CreateBreakpoint(&api.Breakpoint{FunctionName: "main.VPSLLQ36", Line: 4})
 		assertNoError(err, t, "CreateBreakpoint")
 		state := <-c.Continue()
@@ -2817,3 +2783,322 @@ func TestClientServer_SinglelineStringFormattedWithBigInts(t *testing.T) {
 		}
 	})
 }
+
+func TestNonGoDebug(t *testing.T) {
+	// Test that we can at least set breakpoints while debugging a non-go executable.
+	if runtime.GOOS != "linux" {
+		t.Skip()
+	}
+	dir := protest.FindFixturesDir()
+	path := protest.TempFile("testc")
+	cmd := exec.Command("cc", "-g", "-o", path, filepath.Join(dir, "test.c"))
+	if out, err := cmd.CombinedOutput(); err != nil {
+		t.Fatalf("Error compiling %s: %s\n%s", path, err, out)
+	}
+
+	listener, clientConn := service.ListenerPipe()
+	defer listener.Close()
+
+	server := rpccommon.NewServer(&service.Config{
+		Listener:    listener,
+		ProcessArgs: []string{path},
+		Debugger: debugger.Config{
+			Backend:     testBackend,
+			ExecuteKind: debugger.ExecutingExistingFile,
+		},
+	})
+
+	if err := server.Run(); err != nil {
+		t.Fatal(err)
+	}
+
+	client := rpc2.NewClientFromConn(clientConn)
+	defer func() {
+		client.Detach(true)
+	}()
+
+	_, err := client.CreateBreakpoint(&api.Breakpoint{FunctionName: "C.main", Line: -1})
+	if err != nil {
+		t.Fatal(err)
+	}
+}
+
+func TestRestart_PreserveFunctionBreakpoint(t *testing.T) {
+	// Tests that function breakpoint get restored correctly, after a rebuild,
+	// even if the function changed position in the source file.
+
+	dir := protest.FindFixturesDir()
+
+	copy := func(inpath string) {
+		buf, err := os.ReadFile(inpath)
+		assertNoError(err, t, fmt.Sprintf("Reading %q", inpath))
+		outpath := filepath.Join(dir, "testfnpos.go")
+		assertNoError(os.WriteFile(outpath, buf, 0o666), t, fmt.Sprintf("Creating %q", outpath))
+	}
+
+	copy(filepath.Join(dir, "testfnpos1.go"))
+
+	withTestClient2Extended("testfnpos", t, 0, [3]string{}, nil, func(c service.Client, f protest.Fixture) {
+		_, err := c.CreateBreakpoint(&api.Breakpoint{FunctionName: "main.f1"})
+		assertNoError(err, t, "CreateBreakpoint")
+		state := <-c.Continue()
+		assertNoError(state.Err, t, "Continue")
+		t.Logf("%s:%d", state.CurrentThread.File, state.CurrentThread.Line)
+		if state.CurrentThread.Line != 5 {
+			t.Fatalf("wrong location %s:%d", state.CurrentThread.File, state.CurrentThread.Line)
+		}
+
+		// rewrite test file and restart, rebuilding
+		copy(filepath.Join(dir, "testfnpos2.go"))
+		_, err = c.Restart(true)
+		assertNoError(err, t, "Restart(true)")
+
+		state = <-c.Continue()
+		assertNoError(state.Err, t, "Continue")
+		t.Logf("%s:%d", state.CurrentThread.File, state.CurrentThread.Line)
+		if state.CurrentThread.Line != 9 {
+			t.Fatalf("wrong location %s:%d", state.CurrentThread.File, state.CurrentThread.Line)
+		}
+	})
+}
+
+func assertLine(t *testing.T, state *api.DebuggerState, file string, lineno int) {
+	t.Helper()
+	t.Logf("%s:%d", state.CurrentThread.File, state.CurrentThread.Line)
+	if !strings.HasSuffix(state.CurrentThread.File, file) || state.CurrentThread.Line != lineno {
+		t.Fatalf("wrong location %s:%d", state.CurrentThread.File, state.CurrentThread.Line)
+	}
+}
+
+func TestPluginSuspendedBreakpoint(t *testing.T) {
+	if runtime.GOARCH == "ppc64le" {
+		t.Skip("skipped on ppc64le: broken")
+	}
+	// Tests that breakpoints created in a suspended state will be enabled automatically when a plugin is loaded.
+	pluginFixtures := protest.WithPlugins(t, protest.AllNonOptimized, "plugin1/", "plugin2/")
+	dir, err := filepath.Abs(protest.FindFixturesDir())
+	assertNoError(err, t, "filepath.Abs")
+
+	withTestClient2Extended("plugintest", t, protest.AllNonOptimized, [3]string{}, []string{pluginFixtures[0].Path, pluginFixtures[1].Path}, func(c service.Client, f protest.Fixture) {
+		_, err := c.CreateBreakpointWithExpr(&api.Breakpoint{FunctionName: "github.com/go-delve/delve/_fixtures/plugin1.Fn1", Line: 1}, "", nil, true)
+		assertNoError(err, t, "CreateBreakpointWithExpr(Fn1) (suspended)")
+
+		_, err = c.CreateBreakpointWithExpr(&api.Breakpoint{File: filepath.Join(dir, "plugin2", "plugin2.go"), Line: 9}, "", nil, true)
+		assertNoError(err, t, "CreateBreakpointWithExpr(plugin2.go:9) (suspended)")
+
+		cont := func(name, file string, lineno int) {
+			t.Helper()
+			state := <-c.Continue()
+			assertNoError(state.Err, t, name)
+			assertLine(t, state, file, lineno)
+		}
+
+		cont("Continue 1", "plugintest.go", 22)
+		cont("Continue 2", "plugintest.go", 27)
+		cont("Continue 3", "plugin1.go", 6)
+		cont("Continue 4", "plugin2.go", 9)
+	})
+
+	withTestClient2Extended("plugintest", t, protest.AllNonOptimized, [3]string{}, []string{pluginFixtures[0].Path, pluginFixtures[1].Path}, func(c service.Client, f protest.Fixture) {
+		exprbreak := func(expr string) {
+			t.Helper()
+			_, err := c.CreateBreakpointWithExpr(&api.Breakpoint{}, expr, nil, true)
+			assertNoError(err, t, fmt.Sprintf("CreateBreakpointWithExpr(%s) (suspended)", expr))
+		}
+
+		cont := func(name, file string, lineno int) {
+			t.Helper()
+			state := <-c.Continue()
+			assertNoError(state.Err, t, name)
+			assertLine(t, state, file, lineno)
+		}
+
+		exprbreak("plugin1.Fn1")
+		exprbreak("plugin2.go:9")
+
+		// The following breakpoints can never be un-suspended because the
+		// expression is never resolved, but this shouldn't cause problems
+
+		exprbreak("m[0]")
+		exprbreak("*m[0]")
+		exprbreak("unknownfn")
+		exprbreak("+2")
+
+		cont("Continue 1", "plugintest.go", 22)
+		cont("Continue 2", "plugintest.go", 27)
+		cont("Continue 3", "plugin1.go", 5)
+		cont("Continue 4", "plugin2.go", 9)
+	})
+}
+
+// Tests that breakpoint set after the process has exited will be hit when the process is restarted.
+func TestBreakpointAfterProcessExit(t *testing.T) {
+	withTestClient2("continuetestprog", t, func(c service.Client) {
+		state := <-c.Continue()
+		if !state.Exited {
+			t.Fatal("process should have exited")
+		}
+		bp, err := c.CreateBreakpointWithExpr(&api.Breakpoint{ID: 2, FunctionName: "main.main", Line: 1}, "main.main", nil, true)
+		if err != nil {
+			t.Fatal(err)
+		}
+		_, err = c.Restart(false)
+		if err != nil {
+			t.Fatal(err)
+		}
+		state = <-c.Continue()
+		if state.CurrentThread == nil {
+			t.Fatal("no current thread")
+		}
+		if state.CurrentThread.Breakpoint == nil {
+			t.Fatal("no breakpoint")
+		}
+		if state.CurrentThread.Breakpoint.ID != bp.ID {
+			t.Fatal("did not hit correct breakpoint")
+		}
+		if state.CurrentThread.Function == nil {
+			t.Fatal("no function")
+		}
+		if state.CurrentThread.Function.Name() != "main.main" {
+			t.Fatal("stopped at incorrect function")
+		}
+		state = <-c.Continue()
+		if !state.Exited {
+			t.Fatal("process should have exited")
+		}
+		_, err = c.ClearBreakpoint(bp.ID)
+		if err != nil {
+			t.Fatal(err)
+		}
+	})
+}
+
+func TestClientServer_createBreakpointWithID(t *testing.T) {
+	protest.AllowRecording(t)
+	withTestClient2("continuetestprog", t, func(c service.Client) {
+		bp, err := c.CreateBreakpoint(&api.Breakpoint{ID: 2, FunctionName: "main.main", Line: 1})
+		assertNoError(err, t, "CreateBreakpoint()")
+		if bp.ID != 2 {
+			t.Errorf("wrong ID for breakpoint %d", bp.ID)
+		}
+
+		bp2, err := c.CreateBreakpoint(&api.Breakpoint{FunctionName: "main.main", Line: 2})
+		assertNoError(err, t, "CreateBreakpoint()")
+		if bp2.ID != 3 {
+			t.Errorf("wrong ID for breakpoint %d", bp2.ID)
+		}
+	})
+}
+
+func TestClientServer_autoBreakpoints(t *testing.T) {
+	// Check that unrecovered-panic and fatal-throw breakpoints are visible in
+	// the breakpoint list.
+	protest.AllowRecording(t)
+	withTestClient2("math", t, func(c service.Client) {
+		bps, err := c.ListBreakpoints(false)
+		assertNoError(err, t, "ListBreakpoints")
+		n := 0
+		for _, bp := range bps {
+			t.Log(bp)
+			if bp.Name == proc.UnrecoveredPanic || bp.Name == proc.FatalThrow {
+				n++
+			}
+		}
+		if n != 2 {
+			t.Error("automatic breakpoints not found")
+		}
+	})
+}
+
+func TestClientServer_breakpointOnFuncWithABIWrapper(t *testing.T) {
+	// Setting a breakpoint on an assembly function that has an ABI
+	// compatibility wrapper should end up setting a breakpoint on the real
+	// function (also setting a breakpoint on the wrapper is fine).
+	// Issue #3296
+	protest.AllowRecording(t)
+	withTestClient2("math", t, func(c service.Client) {
+		bp, err := c.CreateBreakpoint(&api.Breakpoint{FunctionName: "runtime.schedinit"})
+		assertNoError(err, t, "CreateBreakpoint()")
+		t.Log(bp)
+
+		found := false
+		for _, pc := range bp.Addrs {
+			text, err := c.DisassemblePC(api.EvalScope{}, pc, api.IntelFlavour)
+			assertNoError(err, t, fmt.Sprint("DisassemblePC", pc))
+			t.Log("First instruction for", pc, text[0])
+			if strings.HasSuffix(text[0].Loc.File, "runtime/proc.go") {
+				found = true
+			}
+		}
+		if !found {
+			t.Error("breakpoint not set on the runtime/proc.go function")
+		}
+	})
+}
+
+var waitReasonStrings = [...]string{
+	"",
+	"GC assist marking",
+	"IO wait",
+	"chan receive (nil chan)",
+	"chan send (nil chan)",
+	"dumping heap",
+	"garbage collection",
+	"garbage collection scan",
+	"panicwait",
+	"select",
+	"select (no cases)",
+	"GC assist wait",
+	"GC sweep wait",
+	"GC scavenge wait",
+	"chan receive",
+	"chan send",
+	"finalizer wait",
+	"force gc (idle)",
+	"semacquire",
+	"sleep",
+	"sync.Cond.Wait",
+	"timer goroutine (idle)",
+	"trace reader (blocked)",
+	"wait for GC cycle",
+	"GC worker (idle)",
+	"preempted",
+	"debug call",
+}
+
+func TestClientServer_chanGoroutines(t *testing.T) {
+	protest.AllowRecording(t)
+	withTestClient2("changoroutines", t, func(c service.Client) {
+		state := <-c.Continue()
+		assertNoError(state.Err, t, "Continue()")
+
+		countRecvSend := func(gs []*api.Goroutine) (recvq, sendq int) {
+			for _, g := range gs {
+				t.Logf("\tID: %d WaitReason: %s\n", g.ID, waitReasonStrings[g.WaitReason])
+				switch waitReasonStrings[g.WaitReason] {
+				case "chan send":
+					sendq++
+				case "chan receive":
+					recvq++
+				}
+			}
+			return
+		}
+
+		gs, _, _, _, err := c.ListGoroutinesWithFilter(0, 100, []api.ListGoroutinesFilter{{Kind: api.GoroutineWaitingOnChannel, Arg: "blockingchan1"}}, nil, &api.EvalScope{GoroutineID: -1})
+		assertNoError(err, t, "ListGoroutinesWithFilter(blockingchan1)")
+		t.Logf("blockingchan1 gs:")
+		recvq, sendq := countRecvSend(gs)
+		if len(gs) != 2 || recvq != 0 || sendq != 2 {
+			t.Error("wrong number of goroutines for blockingchan1")
+		}
+
+		gs, _, _, _, err = c.ListGoroutinesWithFilter(0, 100, []api.ListGoroutinesFilter{{Kind: api.GoroutineWaitingOnChannel, Arg: "blockingchan2"}}, nil, &api.EvalScope{GoroutineID: -1})
+		assertNoError(err, t, "ListGoroutinesWithFilter(blockingchan2)")
+		t.Logf("blockingchan2 gs:")
+		recvq, sendq = countRecvSend(gs)
+		if len(gs) != 1 || recvq != 1 || sendq != 0 {
+			t.Error("wrong number of goroutines for blockingchan2")
+		}
+	})
+}
